--[[
	TiltLosses.lua
	
	Author: 	Sushifresser
	Date: 		18.04.2023
	Version: 	0.0.0.1
	
	Changelog:	V 0.0.1.0 @ 18.04.2021 	- initial implementation in FS 22 as extract from WindyLoads mod
]]


TiltLosses = {};
--local TiltLosses_mt = Class(TiltLosses);

local modName = "FS22_TiltLosses";
TiltLosses.MOD_NAME = modName;
local modDirectory = g_currentModDirectory;	--this global variable is only valid while the lua file is being loaded. So we need to store it in our own variable
--TiltLosses.specName = ("spec_%s.TiltLosses"):format(modName);
TiltLosses.currentModDirectory = FS22_TiltLosses.RegisterSpecialization.currentModDirectory;
--TiltLosses.currentModDirectory = modDirectory;

--load the required event files
--source(TiltLosses.currentModDirectory .. "TiltLossEvent.lua");
TiltLosses.modDesc = loadXMLFile("modDesc", TiltLosses.currentModDirectory .. "modDesc.xml");
--some configuration options including their standard values
TiltLosses.DEBUG_MODE = false;	--determines whether debug information will be provided


TiltLosses.ENABLE_TILT_LOSS = true;	--determines whether we want to lose the load when the trailer is tilted
TiltLosses.TILT_START_ANGLE = math.rad(15);	--the angle of the vehicle at which the tilting loss starts. needs to be a value between 0 and 180 degrees
TiltLosses.TILT_END_ANGLE = math.rad(120);	--the angle of the vehicle at which the tilting loss reaches it's maximum. needs to be a value between 0 and 180 degrees and needs to be higher than the start angle
TiltLosses.TILT_COVER_BONUS_ANGLE = math.rad(15);	--the bonus angle is added to the start angle in case the cover is closed to prevent losses on small tilting angles
TiltLosses.TILT_COVER_LOST_ANGLE = math.rad(60);	--this is the angle above which a closed cover cannot hold back the load anymore and it will rip. This is simulated by opening the cover
TiltLosses.TILT_CHECK_CYCLE_MS = 200;	--we only want to check for tilt level losses if the given time has passed to save some calculation load

TiltLosses.modConfigDone = false;	--marker for the completed initial configuration of this mod


TiltLosses.hudData = {
--this determines how the hud will look if it's used
	hudTextPosX = 0.99,
	hudTextPosY = 0.85,
	hudTextSize = 0.015 * g_gameSettings.uiScale
};
TiltLosses.applicableFillTypeCategories = {
--all FillTypeCategories, that are considered for this mod
	"LOADINGVEHICLE",
	"HAYLOFT",
	"WINDROW",
	"COMBINE",
	"FORK",
    "BULK"
};

TiltLosses.excludedFillTypeCategories = {
--all FillTypeCategories which are to be excluded from this mod, even if they are defined in the applicable FillTypeCategories
	"LIQUID",
	"SPRAYER",
	"ANIMAL",
	"HORSE",
	"PRODUCT",
	"SLURRYTANK",
	"PIECE"
};


--this table holds the vent listeners that are used by the TiltLosses mod. We use it to add the event listeners and also remove them if necessary
TiltLosses.USED_EVENT_LISTENERS = {
		"onUpdate",	
		"onDraw" --only for debug hud so far
	};


--this is to store the relation between index and name of a FillType. Can be omitted later to work with the values given by the managers of the game.
--key: (int) fill type index; value: (string) fill type name
TiltLosses.fillTypesIndexToName = {};

--this will store all FillTypes which are considered applicable for this mod
--key: (string) fill type name; value: (table) fillType
TiltLosses.applicableFillTypes = {};


--this is to store vehicle types that are to be excluded from the effects of this mod
--key: (string) full vehicle name; value (boolean) whether the vehicle is to be excluded
TiltLosses.vehicleTypesToExclude = {};

--the constructor of this mod's class
function TiltLosses:new(modDir, isClient)
	-- do this to avoid creating to several instances of the mod
	if g_tiltLossesMod ~= nil then
		return;
	end
	
	-- apply metatable
	--local self = {};
	--setmetatable(self, TiltLosses_mt);

	self.modDir = modDir;
	self.isClient = isClient;

	-- add this instance of TiltLosses Mod as ModEventListener
	--addModEventListener(self);

	return self;
	
end;

--the initialiation routine of this mod. It should be performed only once
function TiltLosses.initModData()
	if g_tiltLossesMod == nil then	--only do the init if it's not already initilaized
		local tiltLossesMod = TiltLosses:new(TiltLosses.currentModDirectory, g_dedicatedServerInfo == nil);
		--local tiltLossesMod = TiltLosses;
		if tiltLossesMod ~= nil then	--did we receive a valid instance of our mod?
			getfenv(0)["g_tiltLossesMod"] = tiltLossesMod;	--create a global reference to our mod instance
			
			--try to find all fillTypes in the game that this mod can be applied to. They will be stored in the mod's class.
			TiltLosses.findApplicableFillTypes();
			--try to read the XML config to configure this mod with user settings
			TiltLosses:getXmlConfig()
			--now we're done
			TiltLosses.modConfigDone = true;
			
			if TiltLosses.DEBUG_MODE then
				print(string.format("[TiltLosses]: Mod directory: %s", tostring(g_modNameToDirectory[TiltLosses.MOD_NAME])));
				print(string.format("[TiltLosses]: Mod successfully initialized with the following parameters: enableTiltLoss: %s", tostring(TiltLosses.ENABLE_TILT_LOSS)));
			end;
				
			delete(TiltLosses.modDesc);
		end;
	end;

end;


--this function is used to init all required mod data
--this function may be called on a vehicle that has been initialized
--this function may also be called on a vehicle that was already initialized. It will be reset and reinitialized then
function TiltLosses.initModOnVehicle(vehicle)
    --vehicle.spec_tiltLosses = vehicle[TiltLosses.specName]
	vehicle.spec_tiltLosses = {};
	
	local spec = vehicle.spec_tiltLosses;
	
	if not TiltLosses.modConfigDone == true then
		TiltLosses.initModData();
	end;
	
	--this stores the last calculated vehicle speed
    spec.lastVehicleSpeed = 0;

	--this table stores all fillUnits of the vehicle, that are susceptible to the mod's effects.
	--key: fillUnitIndex; value: table fillUnit
	spec.supportedFillUnits = {};
	
	--just to be sure, we save the reference to our vehicle here
	spec.vehicle = vehicle;
	
	spec.vehicleSupported, spec.supportedFillUnits = vehicle:getVehicleIsSupportedByTiltLosses(vehicle);

	if TiltLosses.DEBUG_MODE then
		--a table to hold several debug information
		spec.debugInfo = {};
		spec.debugInfo.vehicleWorldRotationX = 0;
		spec.debugInfo.vehicleWorldRotationY = 0;
		spec.debugInfo.vehicleWorldRotationZ = 0;
	end;
	
	spec.modInitialized = true;
	if TiltLosses.DEBUG_MODE then
		print(string.format("[TiltLosses]: Mod successfully loaded for vehicle '%s.'", Vehicle.getFullName(vehicle)));
	end;

end;

function TiltLosses.prerequisitesPresent(specializations)
	local applicable = false;
	applicable = SpecializationUtil.hasSpecialization(FillUnit, specializations) and SpecializationUtil.hasSpecialization(FillVolume, specializations);
	applicable = applicable and SpecializationUtil.hasSpecialization(Dischargeable, specializations);
	return applicable;
end;

function TiltLosses.registerEventListeners(vehicleType)
	for _, functionName in ipairs(TiltLosses.USED_EVENT_LISTENERS) do
		SpecializationUtil.registerEventListener(vehicleType, functionName, TiltLosses);	
		--print(string.format("registered eventlistener %s for vehicle Type %s", functionName, vehicleType.name));
	end;
	--the following event listeners are kept also when we decide to remove the other event listeners if a vehicle is not considered for the mod anymore
	--SpecializationUtil.registerEventListener(vehicleType, "onLoad", TiltLosses);
	SpecializationUtil.registerEventListener(vehicleType, "onLoadFinished", TiltLosses);
	SpecializationUtil.registerEventListener(vehicleType, "onDelete", TiltLosses);	--for removing some stuff
end;

--this function is intended to remove a given vehicle from all the event listeners.
--This can be used for example if we found out that the vehicle cannot be considered for the mod any further
function TiltLosses.removeFromEventListeners(vehicleType)
	for _, functionName in ipairs(TiltLosses.USED_EVENT_LISTENERS) do
		SpecializationUtil.removeEventListener(vehicleType, functionName, TiltLosses);
		if TiltLosses.DEBUG_MODE then
			print(string.format("[TiltLosses]: removed event listener %s for vehicle Type %s", functionName, vehicleType.name));
		end;
	end;
end;

function TiltLosses.registerFunctions(vehicleType)
	SpecializationUtil.registerFunction(vehicleType, "getVehicleIsSupportedByTiltLosses", TiltLosses.getVehicleIsSupportedByTiltLosses);
	SpecializationUtil.registerFunction(vehicleType, "removeLoadFromFillUnit", TiltLosses.removeLoadFromFillUnit);
end;

--this function is called on each vehicle (self) a while loading the savegame
function TiltLosses:onLoadFinished(savegame)
	TiltLosses.initModOnVehicle(self);	--call the initialization routine on this vehicle
end;

--this function is called on each vehicle (self) when the vehicle is deleted
function TiltLosses:onDelete()

end;

--this function is called on each vehicle (self) a while loading the savegame
function TiltLosses:onLoad(savegame)
	TiltLosses.initModOnVehicle(self);	--call the initialization routine on this vehicle
end;

--this function tries to extract some configuration values from the mod description XML file.
function TiltLosses:getXmlConfig()

	local modDesc = TiltLosses.modDesc;
	TiltLosses.DEBUG_MODE = Utils.getNoNil(getXMLBool(modDesc, "modDesc.tiltLossesConfigurations#debugMode"), TiltLosses.DEBUG_MODE);	--we read the debug mode option from the modDesc file
	local specializationNumber = 0;
	--at first we take care of the custom fill type configurations
	while true do
		local specializationKey = string.format("modDesc.tiltLossesConfigurations.fillTypeConfigurations.fillTypeConfiguration(%d)", specializationNumber);
		if not hasXMLProperty(modDesc, specializationKey) then
			break;
		end;
		local fillTypeName = (Utils.getNoNil(getXMLString(modDesc, specializationKey .. "#fillTypeName"), "")):upper();
		if fillTypeName ~= "" then
			local fillTypeInclude = Utils.getNoNil(getXMLBool(modDesc, specializationKey .. "#fillTypeInclude"), false);
			local fillTypeExclude = Utils.getNoNil(getXMLBool(modDesc, specializationKey .. "#fillTypeExclude"), false);
			if fillTypeInclude and fillTypeExclude then
			--only one of the attributes should be provided. We will handle this item as to be excluded.
				Logging.xmlWarning(modDesc, "[TiltLosses]: TiltLosses recieved same fillType '%s' to be included an excluded at the same time. Item will be handled as to be excluded. Please check your configuration!", fillTypeName);
				fillTypeInclude = false;
			end;
			if fillTypeExclude then
				if TiltLosses.applicableFillTypes[fillTypeName] ~= nil then
					TiltLosses.fillTypesIndexToName[TiltLosses.applicableFillTypes[fillTypeName].index] = nil;
					TiltLosses.applicableFillTypes[fillTypeName] = nil;
					if TiltLosses.DEBUG_MODE then
						print(string.format("[TiltLosses]: Removed FillType '%s' from applicable FillTypes of TiltLosses mod based on user-defined XML configuration.", fillTypeName));
					end;
				end
			elseif fillTypeInclude then
				if TiltLosses.applicableFillTypes[fillTypeName] == nil then
				-- the fill type needs to be added at first
					local fillType = g_fillTypeManager:getFillTypeByName(fillTypeName);
					if fillType~= nil and fillType.name ~= nil then
						TiltLosses.applicableFillTypes[fillType.name] = fillType;
						TiltLosses.fillTypesIndexToName[fillType.index] = fillType.name;
						if TiltLosses.DEBUG_MODE then
							print(string.format("[TiltLosses]: Added FillType '%s' to applicable FillTypes of TiltLosses mod based on user-defined XML configuration.", fillType.name));
						end;
					end;
				end;
			end;
		end;
			
		specializationNumber = specializationNumber + 1;
	end;
	specializationNumber = 0;
	-- now take another round for the custom vehicle type configurations
	while true do
		local specializationKey = string.format("modDesc.tiltLossesConfigurations.vehicleTypeConfigurations.vehicleTypeConfiguration(%d)", specializationNumber);
		if not hasXMLProperty(modDesc, specializationKey) then
			break;
		end;
		local vehicleTypeName = (Utils.getNoNil(getXMLString(modDesc, specializationKey .. "#fullVehicleName"), "")):upper();
		if vehicleTypeName ~= "" then
			local vehicleTypeExclude = Utils.getNoNil(getXMLBool(modDesc, specializationKey .. "#vehicleExclude"), false);
			if TiltLosses.vehicleTypesToExclude[vehicleTypeName] == nil then
				TiltLosses.vehicleTypesToExclude[vehicleTypeName] = true; -- now we remember the vehicle and can check at any time we have to
			end;
		end;
		
		specializationNumber = specializationNumber + 1;
	end;
end;

--to get the velocity of the current vehicle
function TiltLosses:getVehicleVelocity()
	local spec = self.spec_tiltLosses;
	--local spec = vehicle[TiltLosses.specName];
	if not spec.modInitialized then
		return;
	end;
	local velocity = self:getLastSpeed();
	if velocity == nil or not velocity or velocity < 0 then
		return 0;
	end
	return velocity;
end;

function TiltLosses:findApplicableFillTypes()
--this function determines all applicable fillTypes based on the definitions of the FillTypeCategories
	local applicableFillTypes = {};
	local fillTypesToCheck = g_fillTypeManager:getFillTypes();
	if TiltLosses.DEBUG_MODE then
		print(string.format("[TiltLosses]: Received " .. #fillTypesToCheck .. " from FillType Manager to check."));
	end;
	--TiltLosses:print_table(fillTypesToCheck, "fillTypesToCheck", "[TiltLosses]: The following fill types have been detected in the game: ");
	for _,fillType in pairs(fillTypesToCheck) do
		if TiltLosses.applicableFillTypes[fillType.name] == nil then
		--only go on if the FillType at hand is not already known
			local isApplicable = false;
			for _, applicableFillTypeCategory in pairs(TiltLosses.applicableFillTypeCategories) do
				if g_fillTypeManager:getIsFillTypeInCategory(fillType.index, applicableFillTypeCategory) then
					isApplicable = true;
					break;
				end;
			end;
			if isApplicable then
			--we found a FillType that seems to be applicable. Now we have to check if it is still to be excluded.
				for _, excludedFillTypeCategory in pairs(TiltLosses.excludedFillTypeCategories) do
					if g_fillTypeManager:getIsFillTypeInCategory(fillType.index, excludedFillTypeCategory) then
						isApplicable = false;
						break;
					end;
				end;
			end;
			if isApplicable then
			--the current fillType has passed all checks and can now be considered applicable
				TiltLosses.applicableFillTypes[fillType.name] = fillType;
				TiltLosses.fillTypesIndexToName[fillType.index] = fillType.name;	-- this table is for more comfortable access to the data
				if TiltLosses.DEBUG_MODE then
					print(string.format("[TiltLosses]: Added FillType '%s' to applicable FillTypes of TiltLosses mod based on standard configuration.", fillType.name));
				end;
			end;
		end;
	end;
	--TiltLosses:print_table(TiltLosses.applicableFillTypes, "TiltLosses.applicableFillTypes", "[TiltLosses]: The following fill types have been registered to be applicable for this mod: ");
end;

---Checks if the vehicle is valid for this mod to be applied to
-- @param table optional vehicle: The reference to the vehicle that is to be examined
-- @return boolean true if the vehicle is supported by this mod. Returns false otherwise
-- @return table Returns table of supported fillUnits of the vehicle if vehicle is supported. Returns nil otherwise.
function TiltLosses:getVehicleIsSupportedByTiltLosses(vehicle)
--this function should check if this vehicle is applicable for this mod
--TODO: maybe find some more useful criteria for checking the vehicles
	-- some vehicle types could be further excluded by default, e.g. combines, balers?
	--local self = vehicle or self;
	local spec = vehicle.spec_tiltLosses;
	--local spec = vehicle[TiltLosses.specName];
	local specFillUnit = vehicle.spec_fillUnit;
	local vehicleSupported = false;
	
	if TiltLosses.vehicleTypesToExclude[Vehicle.getFullName(vehicle):upper()] ~= nil and TiltLosses.vehicleTypesToExclude[Vehicle.getFullName(vehicle):upper()] == true then
		--our list of vehicle types to excludes knows this vehicle type. So we can exit early
		if TiltLosses.DEBUG_MODE then
			print(string.format("[TiltLosses]: Vehicle '%s' is excluded for this mod. It will not be considered any further.", Vehicle.getFullName(vehicle)));
		end;
		TiltLosses.removeFromEventListeners(g_vehicleTypeManager:getTypeByName(vehicle.typeName));
		return false;
	end;
	
	if specFillUnit == nil then	--we obviously need a fill unit in the vehicle for this mod
		if TiltLosses.DEBUG_MODE then
			print("[TiltLosses]: Could not find any FillUnit spec on this vehicle. Vehicle will not be considered for this mod any further");
		end;
		return false;
	end;
	local fillUnits = specFillUnit:getFillUnits();
	if fillUnits == nil or  not (#fillUnits > 0) then
		if TiltLosses.DEBUG_MODE then
			print("[TiltLosses]: Could not find any FillUnits on this vehicle. Vehicle will not be considered for this mod any further");
		end;
		return false;
	end;
	if TiltLosses.DEBUG_MODE then
		print(string.format("[TiltLosses]: Found " .. #fillUnits .. " FillUnits on vehicle '%s'.", Vehicle.getFullName(vehicle)));	
	end;
	
	local specFillVolume = vehicle.spec_fillVolume;	--we also need a fill volume for this mode. Otherwise it may not be an open vehicle which can be affected by wind
	if specFillVolume == nil then
		if TiltLosses.DEBUG_MODE then
			print("[TiltLosses]: Could not find any FillVolume spec on this vehicle. Vehicle will not be considered for this mod any further");
		end;
		return false;
	end;
	
	local specDischargeable = vehicle.spec_dischargeable;
	if specDischargeable == nil then
		if TiltLosses.DEBUG_MODE then
			print("[TiltLosses]: Could not find any Dischargeable spec on this vehicle. Vehicle will not be considered for this mod any further");
		end;
		return false;
	end;
	
	local supportedFillUnits = {};
	for fillUnitIndex, fillUnit in pairs(fillUnits) do
		local fillUnitSupported = false;
		local supportedFillTypes = specFillUnit:getFillUnitSupportedFillTypes(fillUnitIndex);
		if supportedFillTypes ~= nil then
			for fillTypeIndex, _ in pairs(supportedFillTypes) do
			--loop through all supported fillUnitFillTypes to find out if one is included which we can support with our mod
				for fillTypeName, fillType in pairs(TiltLosses.applicableFillTypes) do
					if fillType.index == fillTypeIndex then
						fillUnitSupported = true;
						break;
					end;
				end;
				if fillUnitSupported then
				-- we found the current fillUnit to be supported by our mod. Save it and then go on to the next fillUnit.
					supportedFillUnits[fillUnitIndex] = fillUnit;
					
					spec.tiltLastTimeSync = 0;
					
					vehicleSupported = true;
					break;
				end;
				
			end;
		end;
	end;
	
	if TiltLosses.DEBUG_MODE then
		print(string.format("[TiltLosses]: Vehicle '%s' was %s checked for being applicable for TiltLosses mod.", Vehicle.getFullName(vehicle), (vehicleSupported and "successfully" or "not successfully")));
	end;

	return vehicleSupported, supportedFillUnits;
end;


--this function applies the effects of this mod to all applicable fillUnits of the vehicle
function TiltLosses:affectFillUnits(dt, sendNoEvent)
	local spec = self.spec_tiltLosses;
	--local spec = vehicle[TiltLosses.specName];
	local specFillUnit = self.spec_fillUnit;
	if not spec.modInitialized or spec.vehicle == nil or spec.vehicleSupported == false then
		--there's nothing here for us to compute
		return;
	end;
	
	local vehicleWorldDirection = {};
	vehicleWorldDirection.X, vehicleWorldDirection.Y, vehicleWorldDirection.Z = self:getVehicleWorldDirection();
	if TiltLosses.DEBUG_MODE then
		spec.debugInfo.vehicleWorldDirectionX = vehicleWorldDirection.X or 0;
		spec.debugInfo.vehicleWorldDirectionY = vehicleWorldDirection.Y or 0;
		spec.debugInfo.vehicleWorldDirectionZ = vehicleWorldDirection.Z or 0;
	end;
	for fillUnitIndex, fillUnit in pairs(spec.supportedFillUnits) do
		--we need to check for tilting losses
		spec.tiltLastTimeSync = spec.tiltLastTimeSync + dt;
		if spec.tiltLastTimeSync > TiltLosses.TILT_CHECK_CYCLE_MS then	--check if we have reached the cycle time for checking for tilting losses
			spec.tiltLastTimeSync = 0;
			local remainingVol, angleEffective, realAngle = TiltLosses:getTiltVolumeRemaining(self, fillUnitIndex);
			local currFillLevel = specFillUnit:getFillUnitFillLevelPercentage(fillUnitIndex);
			local tipping = (self.spec_trailer ~= nil and self:getTipState() > 0);	--if we cannot tip, we always calculate the losses
			if tipping == false then --prevent losses when the vehicle is tipping
				--open the cover when the angle is higher than the threshold for the cover to be opened
				if realAngle > TiltLosses.TILT_COVER_LOST_ANGLE then	--check if we want to open the cover because the load bursts out
					if self.spec_cover ~= nil and not specFillUnit:getFillUnitSupportsToolType(fillUnitIndex, ToolType.TRIGGER) then	--this happens when there is a cover that is closed.
						self:setCoverState(fillUnitIndex, false);	--open the cover of the fill unit that is affected
						--TODO: This can be difficult with vehicles that have several covers. Only one would be opened so far.
					end;
				end;
				if (remainingVol) < currFillLevel then
					--here we have to substract some volume from the fill unit because of the tilt level
					if TiltLosses.DEBUG_MODE then
						print(string.format("[TiltLosses]: Need to reduce fill level of fillUnit %d to %.3f %% because of tilt angle of the vehicle", fillUnitIndex, remainingVol*100));
					end;
					TiltLosses.removeLoadFromFillUnit(self, fillUnitIndex, fillUnit.capacity * remainingVol, false);
				end;
			end;
		end;
	end;
end;

--this function is used to actually perform the reduction of the fill unit caused by lost loads
function TiltLosses:removeLoadFromFillUnit(fillUnitIndex, newVol, sendNoEvent)
	local spec = self.spec_tiltLosses;
	--local spec = vehicle[TiltLosses.specName];
	local specFillUnit = self.spec_fillUnit;
	
	local appliedDelta = specFillUnit:addFillUnitFillLevel(self:getOwnerFarmId(), fillUnitIndex, newVol - specFillUnit:getFillUnitFillLevel(fillUnitIndex), specFillUnit:getFillUnitFillType(fillUnitIndex), ToolType.UNDEFINED, nil);

	self:updateMass();

	if not sendNoEvent then
		TiltLossEvent.sendEvent(self, fillUnitIndex, newVol, sendNoEvent);
	end;
end;

--this function calculates the remaining fill volume based on the tilt angles of the vehicle
function TiltLosses:getTiltVolumeRemaining(vehicle, fillUnitIndex)
	local spec = vehicle.spec_tiltLosses;
	--local spec = vehicle[TiltLosses.specName];
	local specFillUnit = vehicle.spec_fillUnit;
	--local specCover = vehicle.spec_cover;
	local vehRotWorld = {};
	local fillUnitIndexLocal = fillUnitIndex or 1;
	--local rootNode = nil;	--this node we want to monitor for the tilt angle
	local rootNode = specFillUnit:getFillUnitAutoAimTargetNode(fillUnitIndex);	--we try to find the fill unit auto aim target node
	if rootNode == nil then	--find another root node if that went wrong
		rootNode = specFillUnit:getFillUnitExactFillRootNode(fillUnitIndex); --the root fill node of the fill unit shall be our root for the emitter of the particles
	end;
	if rootNode == nil then	--find another root node if that also went wrong
		rootNode = vehicle.components[1].node;
	end;
	--vehRotWorld.X, vehRotWorld.Y, vehRotWorld.Z = getWorldRotation(vehicle.components[1].node);
	vehRotWorld.X, vehRotWorld.Y, vehRotWorld.Z = getWorldRotation(rootNode);
	
	if TiltLosses.DEBUG_MODE then	--store some debug information
		spec.debugInfo.vehicleWorldRotationX = vehRotWorld.X;
		spec.debugInfo.vehicleWorldRotationY = vehRotWorld.Y;
		spec.debugInfo.vehicleWorldRotationZ = vehRotWorld.Z;
		spec.debugInfo.tiltRootNode = rootNode;
	end;

	local gamma = math.acos(math.sin(vehRotWorld.X) * math.sin(vehRotWorld.Y) * math.sin(vehRotWorld.Z) + math.cos(vehRotWorld.X) * math.cos(vehRotWorld.Z));
	local bonusAngle = 0;
	if vehicle.spec_cover and not specFillUnit:getFillUnitSupportsToolType(fillUnitIndex, ToolType.TRIGGER) then	--this happens when there is a cover that is closed.
		bonusAngle = TiltLosses.TILT_COVER_BONUS_ANGLE;
	end;
	local angleEffective = math.min(TiltLosses.TILT_END_ANGLE - TiltLosses.TILT_START_ANGLE - bonusAngle, math.max(0, gamma - TiltLosses.TILT_START_ANGLE - bonusAngle)) / (TiltLosses.TILT_END_ANGLE - TiltLosses.TILT_START_ANGLE - bonusAngle) * math.pi/2;
	local remainingVol = 1 - (math.sin(angleEffective)^3);
	remainingVol = math.min(1, math.max(0, remainingVol));	--limit to the absolute boundaries just to be sure
	if TiltLosses.DEBUG_MODE then	--store some debug information
		spec.debugInfo.gamma = gamma;
		spec.debugInfo.angleEffective = angleEffective;
		spec.debugInfo.remainingVol = remainingVol;
	end;
	return remainingVol, angleEffective, gamma;
end;

function TiltLosses:onUpdate(dt, isActiveForInput, isActiveForInputIngnoreSelection, isSelected)
	local spec = self.spec_tiltLosses;
	--local spec = vehicle[TiltLosses.specName];
	if not spec.modInitialized or spec.vehicle == nil or spec.vehicleSupported == false then
		--there's nothing here for us to compute
		return;
	elseif spec.vehicleSupported == nil then
	-- the vehicle should be checked, whether it is applicable for this mod
		spec.vehicleSupported, spec.supportedFillUnits = self:getVehicleIsSupportedByTiltLosses(self);
		if spec.vehicleSupported == false then
			print(string.format("[TiltLosses]: Vehicle '%s' not supported by TiltLosses mod. It will not be considered any further.", Vehicle.getFullName(self)));
			TiltLosses.removeFromEventListeners(g_vehicleTypeManager:getTypeByName(self.typeName));
			return;
		end;
	end;
	--local specCover = self.spec_cover;
	spec.lastVehicleSpeed = TiltLosses.getVehicleVelocity(self);
	TiltLosses.affectFillUnits(self, dt, false);	--we handle all the effects on the fill volume itself
end;

function TiltLosses:onDraw()
	local spec = self.spec_tiltLosses;
	--local spec = vehicle[TiltLosses.specName];
	if not spec.modInitialized or not TiltLosses.DEBUG_MODE or not spec.vehicleSupported then
		return;
	end;
	--print("TiltLosses Draw called");
	if g_currentMission.controlledVehicle ~= nil then
		--only draw debug HUD when inside vehicle
		--local vehicleVel;
		local hudLineLevel = TiltLosses.hudData.hudTextPosY;
		local hudText = "Test String";
		--hudText = string.format("vehicle name: %s", g_currentMission.controlledVehicle:getRootVehicle().name);
		setTextColor(0, 1, 0, 1);
		setTextAlignment(RenderText.ALIGN_RIGHT);
		setTextVerticalAlignment(RenderText.VERTICAL_ALIGN_TOP);
		setTextBold(true);
		if TiltLosses.ENABLE_TILT_LOSS then
			hudText = string.format("vehRot: rX:%.3f; rY:%.3f; rZ:%.3f; gamma: %.1f; eff:%.1f; remVol:%.3f",
				spec.debugInfo.vehicleWorldRotationX,
				spec.debugInfo.vehicleWorldRotationY,
				spec.debugInfo.vehicleWorldRotationZ,
				math.deg(spec.debugInfo.gamma),
				math.deg(spec.debugInfo.angleEffective),
				spec.debugInfo.remainingVol
			);
			renderText(TiltLosses.hudData.hudTextPosX, hudLineLevel, TiltLosses.hudData.hudTextSize, hudText);
			hudLineLevel = hudLineLevel - TiltLosses.hudData.hudTextSize*1.1;
			
			
			--now render the vectors that visualize the angle determination
			local rootNode = {};
			rootNode.x, rootNode.y, rootNode.z = getWorldTranslation(spec.debugInfo.tiltRootNode);
			drawDebugArrow(rootNode.x, rootNode.y, rootNode.z, 0, 2, 0, 0, 0, 0.5, 0, 1, 0);
			local worldQuat = {};
			worldQuat.x, worldQuat.y, worldQuat.z, worldQuat.w = getWorldQuaternion(spec.debugInfo.tiltRootNode);
			--worldQuat.x, worldQuat.y, worldQuat.z, worldQuat.w = mathEulerToQuaternion(spec.debugInfo.vehicleWorldRotationX, spec.debugInfo.vehicleWorldRotationY, spec.debugInfo.vehicleWorldRotationZ);
			local normalVect = {};
			normalVect.x, normalVect.y, normalVect.z = mathQuaternionRotateVector(worldQuat.x, worldQuat.y, worldQuat.z, worldQuat.w, 0, 2, 0);
			drawDebugLine(rootNode.x, rootNode.y, rootNode.z, 1, 0, 0, rootNode.x + normalVect.x, rootNode.y + normalVect.y, rootNode.z + normalVect.z, 1, 0, 0);
		end;
		hudText = string.format("vehicle: v=%.1f km/h@ X=%.3f; Z=%.3f; Y=%.3f",
			spec.lastVehicleSpeed,
			spec.debugInfo.vehicleWorldDirectionX,
			spec.debugInfo.vehicleWorldDirectionZ,
			spec.debugInfo.vehicleWorldDirectionY
			);
			
		
		--renderText(spec.hudData.hudTextPosX, spec.hudData.hudTextPosY, spec.hudData.hudTextSize, "Test Text");
		renderText(TiltLosses.hudData.hudTextPosX, hudLineLevel, TiltLosses.hudData.hudTextSize, hudText);
		
		for fillUnitIndex, fillUnit in pairs(spec.supportedFillUnits) do
			hudLineLevel = hudLineLevel - TiltLosses.hudData.hudTextSize*1.1;
			hudText = string.format("fillUnit [%s]: fillLevel = %.2f", tostring(fillUnitIndex), fillUnit.fillLevel);
            renderText(TiltLosses.hudData.hudTextPosX, hudLineLevel, TiltLosses.hudData.hudTextSize, hudText);
		end;
		
		--print("TiltLosses HUD Text rendered");
		setTextColor(1,1,1,1);
		setTextAlignment(RenderText.ALIGN_LEFT);
		setTextVerticalAlignment(RenderText.VERTICAL_ALIGN_BASELINE);
		setTextBold(false);
	end;
end;

--@end script